/*
 * Scatterer.c
 *
 *  Created on: Oct 13, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */
#include "utilities.h"
#include "Scatterer.h"
#include "Constants.h"
#include "errorHandler.h"
#include "InversionErrorMessages.h"

/**
 * Single instance of scatter list.  There are at most three scatterers.
 * The inversion is solved for one scatterer at a time.  The scatterer
 * is the index into the list and if the value is one, then this is the
 * scatterer of interest.  If the value is zero, then the inversion
 * is not solving for the scatterer.
 *
 * Use get and set operations to access the list.
 *
 * Note: if dim == 2 or 3, then set Scatterer[0] = 1 because when constructing
 * tm matrix, Scatterer[0] is inserted into tm, and it must be one.
 */
static double Scatterer[MAX_ALPHA] = {1.0, 0.0, 0.0};
static char ScattererName[MAX_ALPHA] = {'x', 'y', 'z'};
static int ScattererNum = 0;

/**
 * Construct scatterer list that is zero except for scatterer of interest which is one.
 * @param[in] whichScatterer The scatterer of interest, can be 0, 1 or 2.
 * @return True if valid scatterer; False otherwise.
 */
BOOLEAN setScatterer(int whichScatterer) {
	if ((whichScatterer > 2) || (whichScatterer < 0))
		return FALSE;
	Scatterer[0] = 0.0; // remove default scatterer
	Scatterer[whichScatterer] = 1.0;  // set the one of interest
	ScattererNum = whichScatterer;
	return TRUE;
}

/**
 * Get the list of scatterers.
 * @return Scatterer list.
 */
double *getScatterer() {
	return Scatterer;
}

char translateScatterer(int scattererNum) {
	switch (scattererNum) {
	case 0:
	case 1:
	case 2:
		return ScattererName[scattererNum];
	default:
		reportError(UnrecognizedScatterIndex, "translateScatterer", FATAL);
		return 'E';
	}
}

/**
 * Get the numeric index of the scatterer of interest
 * @return integer representing
 * 0 for x,
 * 1 for y, and
 * 2 for z
 */
int getScatIndex() {
	return ScattererNum;
}

BOOLEAN solvingVZ() {
	return (ScattererNum == 2);
}
