/*
 * Target.c
 *
 *  Created on: Aug 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#include <math.h>
#include <complex.h>
#include "errorHandler.h"
#include "OLABaseErrorMessages.h"
#include "objectfits.h"
#include "matrix.h"
#include "fftw.h"
#include "MapDepthMatrix.h"
#include "ProblemDimension.h"
#include "Gaussian.h"
#include "Scatterer.h"
#include "Constants.h"
#include "BaseDirectory.h"
#include "objectfits.h"
#include "Parameter.h"

#define NORM 1 // switch to choose different normalization schemes for Target function.

/**
 * @defgroup Target Target function - The target function for OLA Averaging Kernels.
 * The target function is implemented as a 2-D Gaussian in x and y, multiplied by a a 1-D function of z.
 * The 1-D function is an input parameter.
 * The Gaussian has an input parameter, sigma_h, which specifies the horizontal
 * full-width at half-maximum (FWHM) of the function.
 *
 * The target function is precalculated for x, y and z values, then Fourier transformed.
 *
 */



///> Single instance of input parameter f(z; z0)
static MATRIX Target_f;

///> Single instance of the target
static MAP_DEPTH_MATRIX *Target = NULL;

struct TargetData {
	int N;
	int Ny;
	int Nz;
	BOOLEAN useFFT;					///< True if using Fourier Transform; False otherwise.
	BOOLEAN debug;
	BaseDirectory *outputBaseDir;
};

static struct TargetData targetData;

/**
 * Verify the dimensions of the target function are compatible with the dimensions of the kernels.
 * @param[in,out] f 1D function f(z) for target function when z > 1.  This parameter is in/out
 * 					because it is Zx1, and if the FITS file is 1xZ, then we can easily take
 * 					the transpose and still run the inversions.
 * @param[in] Nz Size of the z-dimension of the kernels.
 * @returns BOOLEAN Returns TRUE if dimensions are compatible, false otherwise.
 * @ingroup Target
 */
BOOLEAN checkTargetDim(MATRIX *f, int Nz) {
	BOOLEAN ok = TRUE;
	char msg2[MAX_ERROR_MSG_LENGTH];

	// Make sure f.nRow or f.nCol = K.nZ; orient as needed with transpose; f needs to be Zx1
	if (f->nRow != Nz) {
		// try transpose
		if ((f->nCol == Nz) && (f->nRow == 1)) {
			// f is ok, just take the transpose
			MATRIX fT = transpose(*f);
			freeMatrix(*f);
			*f = fT;
		} else {
			ok = FALSE;
			// error, "f needs to have dimensions %d x 1", ZCompare
			snprintf(msg2, MAX_ERROR_MSG_LENGTH, "%s %d x 1", OLABaseTargetfDim, Nz);
			reportError(msg2, "main", ERROR);
		}
	} else if (f->nCol != 1) {
		ok = FALSE;
		// error, "f needs to have dimensions %d x 1", ZCompare
		snprintf(msg2, MAX_ERROR_MSG_LENGTH, "%s %d x 1", OLABaseTargetfDim, Nz);
		reportError(msg2, "main", ERROR);
	}
	return ok;
}



/**
 * Set new parameters for Target function.
 * @param[in] f A MATRIX that contains a 1D function for z-dimension;
 * @param[in] N X-dimension of target function in frequency space.
 * @param[in] Ny Y-dimension in frequency space if useFourierTransform  is true;
 *  different than Nx because of Fourier Transform; otherwise, Y-dimension in real space.
 * @param[in] Nz Z-dimension.
 * @param[in] baseDir Output base directory.
 * @param[in] debug True if running with debug flag; False otherwise.
 * @param[in] useFourierTransform True if transforming target function to wavespace; False otherwise.
 * Multiply 2D Gaussian by this function to get 3D Target function.
 * @ingroup Target
 */
void newTarget(MATRIX f, int N, int Ny, int Nz, BaseDirectory *baseDir, BOOLEAN debug,
		BOOLEAN useFourierTransform) {
// send new value MATRIX f(z,z0) = 1D function
	Target_f = f;
	if (debug) print_matrix2("Target_f", Target_f, ROW_MAJOR);
	targetData.N = N;
	targetData.Ny = Ny;
	targetData.Nz = Nz;
	targetData.useFFT = useFourierTransform;
	targetData.debug = debug;
	targetData.outputBaseDir = baseDir;
}

/**
 * Construct Target function from preset parameters (Gaussian keeps sigma_h,
 * Target_f should already be set to f(z;z0) using newTarget() function).
 * @return Target function stored in a 4-D MAP_DEPTH_MATRIX in wavespace.
 * @ingroup Target
 */
MAP_DEPTH_MATRIX *constructTarget() {
	double value;
	int ix, iy; // indices into matrices
	double x, y; // values for calculating target function
	// because center point is at x = 0, y = 0, z = 0
	// which corresponds to ix=Nx/2, etc.
	MATRIX t;
	BOOLEAN shift = FALSE;
	static MAP_DEPTH_MATRIX T;
	int dim;
	int Nv;
	int Nx, Ny, Nz;
	BOOLEAN debug;

	Nx = targetData.N;
	Ny = targetData.Ny;
	Nz = targetData.Nz;
	debug = targetData.debug;

	dim = getProblemDimension();
	if (dim == 5)
		Nv = 3;
	else
		Nv = 1;

	double *Scatterer = getScatterer();

	// create holding matrix in frequency space coordinates
	if (T.data != NULL)
		destroyMapDepthMatrix(T);
	T = newMapDepthMatrixV(Nx, Ny, Nz, 1, Nv, shift);

	MATRIX fz;
	if (dim == 5) {
		// construct fz (Nz x Nv) x 1 = (Nz x nAlpha) x 1
		// all values are zero except for one scatterer which has Target_fp values;
		if (debug) print_matrix2("Target_f before mult", Target_f, ROW_MAJOR);
		fz = newMatrix(Nz*Nv, 1);
		int a, z;
		double *fzp, *Target_fp;
		fzp = fz.data;
		for (a=0; a<Nv; a++) {
			Target_fp = Target_f.data;
			for (z=0; z<Nz; z++) {
				if (debug) {
					printf("value =%e scatterer[%1d]=%e Target_fp=%e\n",
						Scatterer[a] * *Target_fp, a, Scatterer[a], *Target_fp);
				}
				*fzp = Scatterer[a] * *Target_fp;
				fzp++;
				Target_fp++;
			}
		}
	} else {
		fz = Target_f;
	}

	// Construct T in time space coordinates
	// Assume it is square and use the X-dimension because
	// the Y-dimension is roughly half because of Fourier Transform
	t = newMatrix(Nx, Nx);
	double *p;

	x = (double) -Nx / 2.0;
	y = (double) -Nx / 2.0;

	double C; // The constant C is taken so that the spatial integral of the target function is unity
	C = GaussianConstant();

	double sum = 0.0;
	double sumOfFi = 0.0;

	p = t.data;
	for (ix = 0; ix < Nx; ix++) {
		for (iy = 0; iy < Nx; iy++) {
			value = Gaussian(x + ix, y + iy); // includes C*exp()

			sum += value;
			if (NORM)
				*p++ = value;
			else
				*p++ = C * value * C;

			// multiply by a 1D function of the vertical coordinate, z
			// z0 is an optional parameter and is the depth where the target function is peaked.
			// * f(z; z0) delta(alpha, theta)
		}
	}
	printf("sum of target 2D Gaussian is %e \n", sum);
	if (NORM) {
		MATRIX temp = multiplyScalar(t, 1.0/sum);
		freeMatrix(t);
		t = temp;
	}
	double *fp;
	fp = fz.data;
	int i;
	for (i=0; i<fz.nCol*fz.nRow; i++)
		sumOfFi += fp[i];
	// scale Target_f by 1/sumOfFi, so that the spatial integral of the Target function is one.
	if (sumOfFi == 0.0) {
		reportError("target function is zero", "constructTarget", FATAL);
	}
	MATRIX temp = multiplyScalar(fz, 1.0/sumOfFi);
	if (dim == 5) freeMatrix(fz);
	fz = temp;

	MATRIX tk;
	if (targetData.useFFT) {
		// Fourier transform T
		tk = fftw(MatrixToFitsObject(t));
	} else {
		tk = ConvertToComplex(t);
	}
	insertXYPlaneIntoMapDepthMatrix(tk, 0, 0, &T);
	if (debug) {
		printf("sum of target 1D Fi is 		 %e \n", sumOfFi);
		// write original target function to .fits file before FFT, etc.
		print_matrix2toFile("target", t, ROW_MAJOR);
		print_matrix2toFile("target(k)", tk, ROW_MAJOR);
		print_matrix2("fz", fz, ROW_MAJOR);
	}
	// take care of z-dependence as needed.
	double complex *tkp;
	tkp = tk.data; // this is transformed 2D Gaussian
	if ((tk.nRow != Nx) || (tk.nCol != Ny))
		reportError("Bad dimensions in target function", "constructTarget",
				FATAL);
	for (ix = 0; ix < Nx; ix++) {
		for (iy = 0; iy < Ny; iy++) {
			// extract td; don't extract because the rest of matrix is not set; we only set 0,0
			//MATRIX td = extractMapDepth(ix, iy, T);
			// multiply by f; this creates z-dimension; now dimension should be 1xZ
			MATRIX Tfz = multiplyScalarZ(fz, *tkp);
			// insert td
			insertMapDepth(ix, iy, Tfz, &T);
			freeMatrix(Tfz);
			tkp++;
		}
	}
	freeMatrix(fz);
	freeMatrix(t);

	if (targetData.useFFT) {
		fftw_free(tk.data);  //used fftw_malloc to create this one, so don't free it.
	} else {
		freeMatrix(tk);
	}

	return &T;
}

/**
 * Write target function to output FITS file. Note: the output filename is assumed to be set using
 * newTarget.
 * @param[in] sigmah Current value for sigmah parameter, where sigmah is the full-width at half-maximum (FWHM)
 * of the Gaussian.
 * @param[in] sigmaStep The nth parameter value for sigmah.
 * @ingroup Target
 */
void writeTarget(double sigmah, int sigmaStep) {
	// write the target function
	int NAlpha;
	NAlpha = Target->nV;
	char **filenames;

	filenames = constructOutputFilenames(targetData.outputBaseDir, Constants.TargetFilename, 0, sigmaStep, NAlpha);
	FITSParameters headerParms;
	headerParms.n = 1;
	headerParms.comments[0] = Constants.ModelComment;
	headerParms.names[0] = Constants.ModelSigmaName;
	snprintf(&headerParms.scientificNotation[0][0], MAX_FITS_PARM_VALUE, FITSParametersConstants.sciFiFormat, sigmah);
	writeMapDepthMatrix(filenames, NAlpha, *Target, targetData.useFFT, headerParms);

}

/**
 * Change the target function by recalculating with sigmah.
 * @param[in] sigmah Full-width at half maximum of function.
 * @param[in] sigmaStep The nth sigma step in the parameter range.
 * @ingroup Target
 */
void resetTarget(double sigmah, int sigmaStep) {
	double TwoSqrt2ln2 = 2.0 * sqrt(2.0 * log(2.0));

	double sigma = sigmah / TwoSqrt2ln2;
	newGaussian(0.0, sigma);

	MAP_DEPTH_MATRIX *T;
	T = constructTarget();

	Target = T;
    writeTarget(sigmah, sigmaStep);
}

/**
 * Read the two types of input data for a target function.  First type is a matrix (Nz x 1) containing
 * the z-dimension for the target function. The second type is a complete target function of dimension
 * (Nx x Ny x Nz x Nbeta).
 * @param[in] inputFilename pointer to target file name.
 * @param[in] N Size of Nx dimension of kernels
 * @param[in] Ny Size of Ny dimension of kernels
 * @param[in] Nz Size of Nz dimension of kernels
 * @param[in] sigmaCurrentValue Current value of sigma_h parameter (FWHM).
 * @param[in] sigmaCurrentStep sigmaStep The nth sigma step in the parameter range.
 * @param[in] baseDir pointer to BaseDirectory containing the base directory for all output from the inversion.
 * @param[in] debug  When set to TRUE, intermediate values are written to output directory.
 * @param[in] useFourierTransform TRUE if Fourier transforming the target function on input.
 * @ingroup Target
 */
void readTarget(char *inputFilename, int N, int Ny, int Nz, double sigmaCurrentValue,
		double sigmaCurrentStep, BaseDirectory *baseDir, BOOLEAN debug,
		BOOLEAN useFourierTransform) {
	MATRIX Targetf;
	int dim;
	dim = getProblemDimension();
	if (dim > PDC.D2) {
		if (inputFilename != NULL) {
			// only if z > 1; i.e. 2d problem
			// read the header and determine the dimension 2D or 3D/4D
			MultiDimFits targetInfo;
			targetInfo = openMultiDimFitsRead(inputFilename);
			closeMultiDimFits(targetInfo);

			if (targetInfo.dimension == 2) {
				// if 2D
				Targetf = toMatrix(readfits(inputFilename));
				printf("targetInfo.dimension = %d\n", targetInfo.dimension);
				checkTargetDim(&Targetf, Nz);
			} else {
				// else 3D/4D
				BOOLEAN ok = FALSE;
				printf("targetInfo:\n");
				printf("Nx=%d Ny= %d Nz= %d Nt= %d\n", targetInfo.max.x, targetInfo.max.y,
						targetInfo.max.z, targetInfo.max.t);
				printf("kernelInfo:\n");
				printf("Nx=%d Ny= %d Nz= %d \n", N, Ny, Nz);
                int NyTargetk;   // Ny for target in wave space
                NyTargetk = targetInfo.max.y / 2 + 1;
				if ((targetInfo.max.x == N) && (NyTargetk == Ny) && (targetInfo.max.z == Nz)
						&& (targetInfo.max.t == 1)&& (dim == PDC.D3)) {
					ok = TRUE;
				}
				if ((targetInfo.max.x == N) && (NyTargetk == Ny) && (targetInfo.max.z == Nz)
						&& (dim == PDC.D5) && (targetInfo.max.t == MAX_ALPHA)) {
					ok = TRUE;
				}
				if (ok) {
					static MAP_DEPTH_MATRIX TargetTemp;

					newTarget(newMatrix(0,0), N, Ny, Nz, baseDir, debug, useFourierTransform);
					TargetTemp = readMapDepthMatrixForTarget(inputFilename, useFourierTransform);

					Target = &TargetTemp;
					// Do a little 'fix up' of the target dimensions; since readMapDepthMatrixForTarget
					// puts the scatterer (Nbeta) dimension in nT, we need to move it to nV.
					// The nT (map) dimension for a target function is always 1, because the targets
					// are independent of maps.
					Target->nV = Target->nT;
					Target->nT = 1;

				    // set sigma parameter steps to 1; somehow, check this and send message
					setSigmaStep(1);
					reportError(OLABaseSettingSigmah1, "readTarget", INFO);
					writeTarget(sigmaCurrentValue, sigmaCurrentStep);
				} else {
					char msg2[MAX_ERROR_MSG_LENGTH];
					if (dim == PDC.D3) {
						snprintf(msg2, MAX_ERROR_MSG_LENGTH,
							"Bad dimensions on target function. Expected (Nx,Ny,Nz) (%1d,%1d,%1d); Found (%1d,%1d,%1d)",
							N, (Ny-1)*2, Nz, targetInfo.max.x, targetInfo.max.y, targetInfo.max.z);
					} else {
						snprintf(msg2, MAX_ERROR_MSG_LENGTH,
							"Bad dimensions on target function. Expected (Nx,Ny,Nz,Nbeta) (%1d,%1d,%1d,%1d); Found (%1d,%1d,%1d,%1d)",
							N, (Ny-1)*2, Nz, MAX_ALPHA, targetInfo.max.x, targetInfo.max.y, targetInfo.max.z, targetInfo.max.t);
					}
					reportError(msg2, "readTarget", FATAL);
				}
			}
		} else {
			reportError(OLABaseMissing1dTargetFunc, "readTarget", FATAL);
		}
	} else {
		Targetf = newMatrix(1,1);
		double *p;
		p = Targetf.data;
		p[0] = 1.0;
	}
	// initialize the target function
	// use Fourier transform = TRUE
	if (Target == NULL) {
		newTarget(Targetf, N, Ny, Nz, baseDir, debug, useFourierTransform);
		resetTarget(sigmaCurrentValue, sigmaCurrentStep);

	}
}

/**
 * Returns a pointer to target function represented by a Map-Depth matrix.
 * @returns Pointer to MAP_DEPTH_MATRIX representing the generated target function.
 * @ingroup Target
 */
MAP_DEPTH_MATRIX *getTarget() {
	return Target;
}

/**
 * Free space associated with target function.
 * @ingroup Target
 */
void freeTarget() {
	freeMatrix(Target_f);
	destroyMapDepthMatrix(*Target);
}

