/*
 * main.c
 *
 *  Created on: Aug 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#include <string.h>
#include <math.h>
#include "utilities.h"
#include "Constants.h"
#include "errorHandler.h"
#include "coramfVersion.h"
#include "parsVersion.h"
#include "invrsVersion.h"
#include "obaseVersion.h"
#include "Inversion.h"
#include "OLAInit.h"
#include "OLAInvert.h"
#include "ParameterLoop.h"
#include "ProblemDimension.h"

/**
 * \page InvTypesPage Inversions
 * <B> Subtractive optimally localized averaging inversion (SOLA) </B>
 *
 * @see <a href="../OLAnotes.pdf">Equations and specifics</a>
 * @see OLAMain.c
 * @see OLAInversion
 *
 *
 */

/**
 * Performs OLA inversions on 2D, 3D scalar and 3D vector data.
 * @param[in] inputParms InputParms structure which contains input file names, directories, and
 * parameters for the inversion.
 * @param[in] commandLineOutputDir Character string to override the output base
 * directory specified in the input file.  This parameter is used to create successive
 * versions of the output base directory.
 * @param[in] debug Boolean flag signaling the output of intermediate matrix values.
 * @returns 0 for success and 1 for error.
 */
int run(InputParms inputParms, char *commandLineOutputDir, BOOLEAN debug) {

	static struct Inversion Ola = {
			NULL, // inversion Data
			// define function pointers
			&OLAInit,
			&OLAInvert2,
			&OLAComplete,
			&OLADestroy,
			&OLAMoreParameterSteps,
			&OLANextParameter
	};

	printf("Inversion %s\n", CODE_VERSION);
	printf("Libraries %s\n", getLibraryVersion());
	printf("coramf Version %s\n", coramfVersion);
	printf("pars Version %s\n", parsVersion);
	printf("invrs Version %s\n", invrsVersion);
	printf("obase Version %s\n", obaseVersion);
	// check library and inversion for compatible versions
	if ((strcmp(CODE_VERSION, getLibraryVersion()) != 0) ||
			(strcmp(CODE_VERSION, coramfVersion) != 0) ||
			(strcmp(CODE_VERSION, parsVersion) != 0) ||
			(strcmp(CODE_VERSION, obaseVersion) != 0) ||
			(strcmp(CODE_VERSION, invrsVersion) != 0))
		reportError(BadCodeVersions, "main", FATAL);

	struct Inversion op;

	if (inputParms.ola) {
		op = Ola;
		printf("using ola\n");
	} else
		reportError("Unrecognized type of inversion", "main", FATAL);

	op.data = op.init(inputParms, commandLineOutputDir, debug);

	// loop over parameters
	while (op.moreParameters(op.data)) {
		op.solve(op.data);
		op.complete(op.data);
		op.nextParameter(op.data);
	}
	// end loop
	op.destroy(op.data);

	return(EXIT_SUCCESS);

}
